/* 
 * Copyright (C) 2009-2010 MediaTek Co.,Ltd.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 */

#include <linux/init.h>
#include <linux/module.h>
#include <linux/platform_device.h>
#include <linux/input.h>
#include <mach/upmu_common.h>
#include <mach/upmu_common_sw.h>
#include <mach/upmu_sw.h>
#include <mach/upmu_hw.h>

#define mtk_vib_haptic_enable()  pmic_set_register_value(PMIC_RG_VIBR_EN, 1)
#define mtk_vib_haptic_disable() pmic_set_register_value(PMIC_RG_VIBR_EN, 0)

struct mtk_vib_data {
    struct input_dev *input_dev;
    struct device *dev;
    struct work_struct work;
    int speed;
};

static void mtk_vib_close(struct input_dev *dev)
{
    struct mtk_vib_data *vib = input_get_drvdata(dev);
    cancel_work_sync(&vib->work);
    mtk_vib_haptic_disable();
}

static void mtk_vib_play_effect_work(struct work_struct *work)
{
    struct mtk_vib_data *vib = container_of(work, struct mtk_vib_data, work);

    if(vib->speed)
        mtk_vib_haptic_enable();
    else
        mtk_vib_haptic_disable();
}

static int mtk_vib_play_effect(struct input_dev *dev, void *data,
        struct ff_effect *effect)
{
    struct mtk_vib_data *vib = input_get_drvdata(dev);

    vib->speed = effect->u.rumble.strong_magnitude;
    if (!vib->speed)
        vib->speed = effect->u.rumble.weak_magnitude;

    schedule_work(&vib->work);
    return 0;
}

static struct mtk_vib_data mtk_vibrator_data;
static int mtk_vib_probe(struct platform_device *pdev)
{
    int ret;
    struct mtk_vib_data *vib;

    platform_set_drvdata(pdev, &mtk_vibrator_data);
    vib = (struct mtk_vib_data *)platform_get_drvdata(pdev);
    pdev->dev.init_name = "mtk_vibrator";

    INIT_WORK(&vib->work, mtk_vib_play_effect_work);

    vib->input_dev = input_allocate_device();
    if (!vib->input_dev) {
        dev_err(&pdev->dev, "unable to allocate memory\n");
        ret = -ENOMEM;
        goto err_free_mem;
    }

    vib->input_dev->name = "mtk_vibrator";
    vib->input_dev->id.version = 1;
    vib->input_dev->close = mtk_vib_close;
    input_set_drvdata(vib->input_dev, &mtk_vibrator_data);
    input_set_capability(vib->input_dev, EV_FF, FF_RUMBLE);

    ret = input_ff_create_memless(vib->input_dev, NULL, mtk_vib_play_effect);
    if (ret) {
        dev_err(&pdev->dev, "unable to create FF device, error: %d\n", ret);
        goto err_free_mem;
    }

    ret = input_register_device(vib->input_dev);
    if (ret) {
        dev_err(&pdev->dev, "unable to register input device, error: %d\n", ret);
        goto err_destroy_ff;
    }
    return 0;

err_destroy_ff:
    input_ff_destroy(vib->input_dev);
err_free_mem:
    input_free_device(vib->input_dev);
    return ret;
}

static int mtk_vib_remove(struct platform_device *pdev)
{
    struct mtk_vib_data *vib = platform_get_drvdata(pdev);

    input_unregister_device(vib->input_dev);
    mtk_vib_haptic_disable();

    return 0;
}

static void mtk_vib_shutdown(struct platform_device *pdev)
{
    mtk_vib_haptic_disable();
}

static struct platform_driver mtk_vib_driver = {
    .probe = mtk_vib_probe,
    .remove = mtk_vib_remove,
    .shutdown = mtk_vib_shutdown,
    .driver = {
        .name = "mtk_vibrator",
        .owner = THIS_MODULE,
    },
};

static struct platform_device mtk_vib_device = {
    .name = "mtk_vibrator",
    .id = -1,
};

static int mtk_vib_init(void)
{
    int ret;

    /*
     * Allowed voltage for PMIC_RG_VIBR_VOSEL value:
     * 0x0: 1.3V  0x1: 1.5V  0x2: 1.8V  0x3: 2.0V
     * 0x4: 2.5V  0x5: 2.8V  0x6: 3.0V  0x7: 3.3V
     */
    pmic_set_register_value(PMIC_RG_VIBR_VOSEL, 0x6);

    ret = platform_device_register(&mtk_vib_device);
    if (ret != 0) {
        printk("[vibrator] Unable to register vibrator device (%d)\n", ret);
        return ret;
    }

    ret = platform_driver_register(&mtk_vib_driver);
    if (ret) {
        printk("[vibrator] Unable to register vibrator driver (%d)\n", ret);
        return ret;
    }

    return 0;
}

static void mtk_vib_exit(void)
{
    platform_driver_unregister(&mtk_vib_driver);
    platform_device_unregister(&mtk_vib_device);
}

module_init(mtk_vib_init);
module_exit(mtk_vib_exit);
MODULE_AUTHOR("MediaTek Inc.");
MODULE_DESCRIPTION("MTK Vibrator Driver (VIB)");
MODULE_LICENSE("GPL");
